/*
 * Copyright (C) 2001-2006 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */

import mir.bundle.Bundle;
import mir.config.MirPropertiesConfiguration;
import mir.servlet.AbstractServlet;
import mir.servlet.AdminServletModule;
import mir.servlet.ServletModuleExc;
import mir.servlet.ServletModuleUserExc;
import mir.util.ExceptionRoutines;
import mir.util.StringRoutines;
import mircoders.entity.EntityUsers;
import mircoders.global.MirGlobal;
import mircoders.module.ModuleUsers;
import mircoders.servlet.ServletHelper;
import multex.Failure;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.UnavailableException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionBindingListener;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;

public class Mir extends AbstractServlet {
  private static ModuleUsers usersModule = null;
  private static Locale fallbackLocale = null;

  private static List loginLanguages = null;

  private List getLoginLanguages() throws Failure {
    synchronized (Mir.class) {
      try {
        if (loginLanguages == null) {
          List languages =
            StringRoutines.splitString(MirGlobal.config().getString("Mir.Login.Languages", "en"), ";");

          loginLanguages = new ArrayList();

          Iterator i = languages.iterator();

          while (i.hasNext()) {
            String code = (String) i.next();

            Bundle bundle =
                MirGlobal.getBundleFactory().getBundle("etc/bundles/adminlocal", new String[] { code });
            Bundle defaultBundle =
                MirGlobal.getBundleFactory().getBundle("bundles/admin", new String[] { code });

            String name = bundle.getValue("languagename", Collections.EMPTY_LIST);

            if (name == null) {
              name = defaultBundle.getValue("languagename", Collections.EMPTY_LIST);
            }

            if (name == null) {
              name = code;
            }

            Map record = new HashMap();
            record.put("name", name);
            record.put("code", code);
            loginLanguages.add(record);
          }
        }

        return loginLanguages;
      }
      catch (Throwable t) {
        throw new Failure("Error while retrieving the available login languages", t);
      }
    }
  }

  public void init(ServletConfig config) throws ServletException {
    super.init(config);

    usersModule = new ModuleUsers();
  }

  protected String getDefaultLanguage(HttpServletRequest aRequest) {
    String defaultlanguage =
      MirGlobal.config().getString("Mir.Login.DefaultLanguage", "");

    if (defaultlanguage.length() == 0) {
      Locale locale = aRequest.getLocale();
      defaultlanguage = locale.getLanguage();
    }

    return defaultlanguage;
  }

  protected synchronized Locale getFallbackLocale() throws ServletException {
    try {
      if (fallbackLocale == null) {
        fallbackLocale = new Locale(MirPropertiesConfiguration.instance().getString("Mir.Admin.FallbackLanguage", "en"), "");
      }
    }
    catch (Throwable t) {
      throw new ServletException(t.getMessage());
    }

    return fallbackLocale;
  }

  public EntityUsers checkCredentials(HttpServletRequest aRequest) throws ServletException {
    try {
      EntityUsers user = ServletHelper.getUser(aRequest);

      String username = aRequest.getParameter("login");
      String password = aRequest.getParameter("password");

      if (username != null && password != null) {
        user = usersModule.getUserForLogin(username, password);

        if (user!=null) {
          ServletHelper.setUser(aRequest, user);
          usersModule.recordLogin(user);
          aRequest.getSession().setAttribute("sessiontracker", new SessionTracker(username, user.getId()));
        }
      }

      return user;
    }
    catch (Throwable t) {
      throw new ServletException(t.getMessage());
    }
  }

  public void process(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletException, IOException, UnavailableException {
    try {
      long startTime = System.currentTimeMillis();
      long sessionConnectTime = 0;

      HttpSession session = aRequest.getSession(true);
      setNoCaching(aResponse);
      aResponse.setContentType("text/html; charset=" +
                               configuration.
                               getString("Mir.DefaultHTMLCharset", "UTF-8"));

      EntityUsers userEntity = checkCredentials(aRequest);

      if (userEntity == null) {
        String queryString = aRequest.getQueryString();

        if ( (queryString != null) && (queryString.length() != 0) && session.getAttribute("login.target") == null &&
             (aRequest.getParameter("module")==null ||
              (!aRequest.getParameter("module").equals("login") && !aRequest.getParameter("module").equals("logout")))) {
          session.setAttribute("login.target", queryString);
        }

        _sendLoginPage(aResponse, aRequest);
      }
      else {
        String moduleName = aRequest.getParameter("module");
        checkLanguage(session, aRequest);

        if ( ( (moduleName == null) || moduleName.equals(""))) {
          moduleName="Admin";
        }

        if (moduleName.equals("login")) {
          String target = (String) session.getAttribute("login.target");

          if (target != null) {
            ServletHelper.redirect(aResponse, target);
          }
          else {
            ServletHelper.redirect(aResponse, "");
          }
        }
        else if (moduleName.equals("logout")) {
          logger.info(userEntity.getFieldValue("login") + " has logged out");
          session.invalidate();
          _sendLoginPage(aResponse, aRequest);
          return;
        }
        else {
          try {
            AdminServletModule servletModule = getServletModuleForName(moduleName);
            servletModule.handleRequest(aRequest, aResponse);

            sessionConnectTime = System.currentTimeMillis() - startTime;
            logger.info("EXECTIME (" + moduleName + "): " + sessionConnectTime + " ms");
          }
          catch (Throwable e) {
            Throwable cause = ExceptionRoutines.traceCauseException(e);

            if (cause instanceof ServletModuleUserExc)
              handleUserError(aRequest, aResponse, (ServletModuleUserExc) cause);
            else
              handleError(aRequest, aResponse, cause);
          }

          if (aRequest.getParameter("killsession")!=null)
            aRequest.getSession().invalidate();
        }
      }
    }
    catch (Throwable t) {
      throw new ServletException(t.toString());
    }
  }

  /**
   * caching routine to get a module for a module name
   *
   * @param aModuleName the module name
   * @return the requested module
   * @throws ServletModuleExc
   */
  private static AdminServletModule getServletModuleForName(String aModuleName) throws ServletModuleExc {
    return ServletHelper.getServletModule(aModuleName);
  }

  private void handleUserError(HttpServletRequest aRequest, HttpServletResponse aResponse, ServletModuleUserExc anException) {
    try {
      logger.info("user error: " + anException.getMessage());

      Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, new Locale[] {getLocale(aRequest), getFallbackLocale()});

      Bundle bundle =
          MirGlobal.getBundleFactory().getBundle("etc/bundles/adminlocal", new
              String[] { getLocale(aRequest).getLanguage() });
      Bundle defaultBundle =
          MirGlobal.getBundleFactory().getBundle("bundles/admin", new
              String[] { getLocale(aRequest).getLanguage() });
      String message =
        bundle.getValue(anException.getMessage(), Arrays.asList(anException.getParameters()));

      if (message==null) {
        message =
          defaultBundle.getValue(anException.getMessage(), Arrays.asList(anException.getParameters()));
      }

      responseData.put("errorstring", message);
      responseData.put("date", new GregorianCalendar().getTime());

      ServletHelper.generateResponse(aResponse.getWriter(), responseData, MirPropertiesConfiguration.instance().getString("Mir.UserErrorTemplate"));
    }
    catch (Throwable e) {
      logger.error("Error handling user error" + e.toString());
    }
  }

  private void handleError(HttpServletRequest aRequest, HttpServletResponse aResponse, Throwable anException) {
    try {
      logger.error("error: " + anException);

      Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, new Locale[] {getLocale(aRequest), getFallbackLocale()});

      responseData.put("errorstring", anException.toString());
      StringWriter writer = new StringWriter();
      anException.printStackTrace(new PrintWriter(writer));
      responseData.put("stacktrace", writer.toString());
      responseData.put("date", new GregorianCalendar().getTime());

      ServletHelper.generateResponse(aResponse.getWriter(), responseData, MirPropertiesConfiguration.instance().getString("Mir.ErrorTemplate"));
    }
    catch (Throwable e) {
      logger.error("Error handling error: " + e.toString());

      try {
        Throwable rootException = ExceptionRoutines.traceCauseException(anException);

        PrintWriter writer = aResponse.getWriter();
        writer.println("<html><head><title>FATAL Error</title><body>");
        writer.println("<h1>" + rootException.toString()+"</h1>");
        writer.println("<code>");
        rootException.printStackTrace(writer);
        writer.println("</code>");
        writer.println("</body></html>");
        writer.close();
      }
      catch (Throwable t) {

      }
    }
  }

  // Redirect-methods
  private void _sendLoginPage(HttpServletResponse aResponse, HttpServletRequest aRequest) {
    String loginTemplate = configuration.getString("Mir.LoginTemplate");

    try {
      Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, new Locale[] {getLocale(aRequest), getFallbackLocale()});

      responseData.put("defaultlanguage", getDefaultLanguage(aRequest));
      responseData.put("languages", getLoginLanguages());

      ServletHelper.generateResponse(aResponse.getWriter(), responseData, loginTemplate);
    }
    catch (Throwable e) {
      handleError(aRequest, aResponse, e);
    }
  }

  public String getServletInfo() {
    return "Mir " + configuration.getString("Mir.Version");
  }

  private class SessionTracker implements HttpSessionBindingListener {
    private String name;
    private String id;

    public SessionTracker(String aUserName, String anId) {
      name = aUserName;
      id = anId;
    }

    public void valueBound(HttpSessionBindingEvent anEvent) {
      MirGlobal.registerLogin(name, id);
    }

    public void valueUnbound(HttpSessionBindingEvent anEvent) {
      MirGlobal.registerLogout(name, id);
    }
  }
}
